from matplotlib.colors import LinearSegmentedColormap as LSC
from math import pi, floor
import numpy as np
from colorspacious import cspace_converter

def cmap(start=15, rot=-1.5, gamma=1.0, gamp=1.0, grot=1.0, ginv=1.0,
         pwr_amp=1.0, eps_amp=0.05, eps_pwr=2, sw_sc=0, pwr_sc=2,
         k_ellipse=0.0, alpha_ellipse=0.0, insectoid=False,
         reverse=False, invmode=0, cuberot=False, nlev=256, minSat=1.,
         maxSat=1., scalsat=1., minLight=0., maxLight=1., scallight=1,
         ColorModel=2, iterSat=100, epsSat=1.e-6, cliprange=[0.1, 0.9],
         wluma=[0.3, 0.59, 0.11], trimWhite=True, verbose=0, **kwargs):
    """
    A full implementation of Dave Green's "cubehelix" extended by Ingo Thies
    for Matplotlib and Colorspacious.
    Based on cubehelix.py by James R. A. Davenport and contributors
    which itself is based on the FORTRAN 77 code provided in
    D.A. Green, 2011, BASI, 39, 289.

    http://adsabs.harvard.edu/abs/2011arXiv1108.5083G

    User can adjust all parameters of the cube/colorhelix algorithm.
    This enables much greater flexibility in choosing color maps, while
    always ensuring the color map scales in intensity from black
    to white. A few simple examples:

    Default color map settings produce the standard "cubehelix".

    Create color map in only blues by setting rot=0 and start=0.

    Create reverse (white to black) backwards through the rainbow once
    by setting rot=1 and reverse=True.

    Extension to colorhelix allows for use of the state-of-the-art color space
    CIECAM02 and its UCS variant.

    Requires the colorspacious module to be installed.

    Parameters
    ----------
    start : scalar, optional
        Sets the starting position in the color space in degrees.
        15=red, 105=yellow, 135=green, 270=blue. Defaults to 60
    rot : scalar, optional
        The number of rotations through the rainbow. Can be positive
        or negative, indicating direction of rainbow. Negative values
        correspond to Blue->Red direction. Defaults to -1.5
    gamma : scalar, optional
        The gamma correction for intensity. Defaults to 0.8 (good for CAM02-UCS)
    gamp : scalar, optional
        The gamma correction for chromaticity variable amp. Defaults to 1.0
    grot : scalar, optional
        The gamma correction for rotation. Defaults to 1.0
    pwr_amp : scalar, optional
        Additional exponent to chromaticity amp. Defaults to 1.0
    eps_amp : scalar, optional
        Smoothing length used for negative pwr_amp only. Defaults to 0.05
    eps_pwr : scalar, optional
        Smoothing power used for negative pwr_amp only. Defaults to 2
    sw_sc : scalar, optional
        S-curve distortion for lightness. >0: flattens, <0:steepens in the middle.
        Defaults to 0
    pwr_sc : scalar, optional
        Strength exponent for s-curve distortion. Defaults to 2.
    k_ellipse : scalar, optional
        Elliptical deformation parameter. 0=off, 1=circle, 0<k<1 ratio of
        major to minor axis. Default to 0 (off)
    alpha_ellipse : scalar, optional
        Orientation of ellipse major axis. Defaults to 0
    insectoid : boolean, optional
        Toggle variant with insectoid-shape instead of elliptcal deformation.
        Defaults to False
    reverse : boolean, optional
        Set to True to reverse the color map. Will go from black to
        white. Good for density plots where shade~density. Defaults to False
    cuberot : boolean, optional
        Set to True to enable legacy cubehelix rotation scaling (0-3).
        Defaults to False
    nlev : scalar, optional
        Defines the number of discrete levels to render colors at.
        Defaults to 256.
    sat : scalar, optional
        The saturation intensity factor. Defaults to 1
        NOTE: this was formerly known as "hue" parameter
    minSat : scalar, optional
        Sets the minimum-level saturation. Defaults to 1
    maxSat : scalar, optional/
        Sets the maximum-level saturation. Defaults to 1
    scalsat : scalar, optional
        Scales the optimum saturation. Note that values > 1 may cause a chunky
        appearance. Defaults to 1
    startHue : scalar, optional
        Sets the starting color, ranging from [0, 360], as in
        D3 version by @mbostock
        NOTE: overrides values in start parameter
    endHue : scalar, optional
        Sets the ending color, ranging from [0, 360], as in
        D3 version by @mbostock
        NOTE: overrides values in rot parameter
    minLight : scalar, optional
        Sets the minimum lightness value. Defaults to 0
    maxLight : scalar, optional
        Sets the maximum lightness value. Defaults to 1
    scallight : scalar, optional
        Scales the overall lightness value. Defaults to 1
    ColorModel : integer, optional
        Source color space used. Defaults to 0
        0: Classic cubehelix
        1: Standard CIECAM02 (with standard viewing conditions)
        2: CIECAM02-UCS (Luo et al. 2006)
    iterSat : integer, optional
        Maximum iteration number for maximising no-clipping saturation. 0=off.
        Defaults to 100
    epsSat : scalar, optional
        Accuracy criterion for saturation iteration. Defaults to 1e-6
    cliprange : 1D array-like, lenght 2, optional
        Range limits in intensity inside which clipping is tested for saturation iteration.
        Defaults to [0.1, 0.9]
    wluma : array-like, optional
        Sets the R,G,B luma weights for cubehelix only. Defaults to [0.3, 0.59, 0.11]
    trimWhite : boolean, optional
        Forces the upper endpoint to be white (which may not be exactly the case for
        CIECAM02 models). No effect for non-white endpoint or cubehelix.
        Defaults to False
    tintMode : integer, optional
        Toggles endpoint tinting.
    tintGamma : scalar, optional
        adjusts neutral gray for tintMode==2
    startRGB : array, optional
        start (dark side) tint RGB
    endRGB : array, optional
        end (bright side) tint RGB
    startC3 : array, optional
        start (dark side) scale RGB (tintMode==2)
    endC3 : array, optional
        end (bright side) scale RGB (tintMode==2)
    startPwr : scalar, optional
        start (dark side) tint power
    endPwr : scalar, optional
        end (bright side) tint power
    whtPwr : scalar, optional
        end (bright side) white convergence power. Used if > 0.
    swwht : scalar, optional
        end (bright side) white convergence mixer.
    verbose : scalar, optional
        Sets the verbosity level. verbose > 1 gives additional output.
        verbose == 1 is reserved for optional calling optimisation function.

    Returns
    -------
    matplotlib.colors.LinearSegmentedColormap object

    Example
    -------
    >>> import colorhelix
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> x = np.random.randn(1000)
    >>> y = np.random.randn(1000)
    >>> cx = colorhelix.cmap(start=0., rot=-0.5)
    >>> plt.hexbin(x, y, gridsize=50, cmap=cx)
    """
    global auxpars

    auxpars = []

# cubehelix rotation mode
    if cuberot: # scale rot to degrees
        start *= 120.
    elif ColorModel <= 0: # add 90 degrees offset
        start += 90.
    do_tint = False
    rgbtint0 = np.array([1, 1, 1])
    rgbtint1 = np.array([1, 1, 1])
    scalrgb0 = np.array([1, 1, 1])
    scalrgb1 = np.array([1, 1, 1])
    pwrtint0 = 2
    pwrtint1 = 2
    pwrtintw = 4
    swmixw = 1
    gtint = gamma

# override start and rot if startHue and endHue are set
    if kwargs is not None:
        if 'startHue' in kwargs:
            start = kwargs.get('startHue')
        if 'endHue' in kwargs:
            rot = (kwargs.get('endHue') - start) / 360.
        if 'sat' in kwargs:
            minSat = kwargs.get('sat')
            maxSat = kwargs.get('sat')
        if 'tintMode' in kwargs:
            tintmode = kwargs.get('tintMode')
            do_tint = 0 < tintmode < 3
        if 'tintGamma' in kwargs:
            gtint = kwargs.get('tintGamma')
            if gtint <= 0:
                gtint = gamma
        if 'startRGB' in kwargs:
            rgbtint0 = kwargs.get('startRGB')
            rgbtint0 = cspace_converter("sRGB1", "sRGB1-linear")(rgbtint0)
        if 'endRGB' in kwargs:
            rgbtint1 = kwargs.get('endRGB')
            rgbtint1 = cspace_converter("sRGB1", "sRGB1-linear")(rgbtint1)
        if 'startC3' in kwargs:
            scalrgb0 = kwargs.get('startC3')
            print('scalrgb0 = ',scalrgb0)
        if 'endC3' in kwargs:
            scalrgb1 = kwargs.get('endC3')
        if 'startPwr' in kwargs:
            pwrtint0 = kwargs.get('startPwr')
        if 'endPwr' in kwargs:
            pwrtint1 = kwargs.get('endPwr')
        if 'whtPwr' in kwargs:
            pwrtintw = kwargs.get('whtPwr')
        if 'swwht' in kwargs:
            swmixw = kwargs.get('swwht')

    if verbose > 1:
        endhue = start+360.*rot
        endhue -= 360*floor(endhue/360)
        print('End angle = {:.2f} deg'.format(endhue))

# set up the parameter
    fract_full = np.linspace(0, 1, nlev)
    #fract_part = np.linspace(minLight, maxLight, nlev)
    fract_part = np.linspace(max(minLight, 1.e-15), max(maxLight, 1.e-15), nlev) #avoid math warning from colorspacious/ciecam02.py
    if scallight > 0:
        fract_l = scallight*scurve(fract_part, sw_sc, pwr_sc)**gamma
    else:
        fract_l = scurve(fract_part, sw_sc, pwr_sc)**gamma
    fract_r = fract_full**grot
    fract_a = fract_part**gamp
    
    angle = 2 * pi * (start / 360. + rot * fract_r)

    alpha_ellipse *= pi/180.

    if ColorModel <= 0:
# compute the RGB base vectors from luma weights
        wr, wg, wb = wluma
        sumw = wr+wg+wb
        wr /= sumw
        wg /= sumw
        wb /= sumw
        rho_rg = np.sqrt(wr**2+wg**2)
        V = np.array([wg/rho_rg, -wr/rho_rg, 0.])
        eps0 = wb*wr/(wr**2+wg**2)
        u_norm = np.sqrt(1.+eps0**2+(wg/wr*eps0)**2)
        eps = eps0/u_norm
        U = np.array([-eps, -eps*wg/wr, 1./u_norm])
        cang = np.cos(angle)
        sang = np.sin(angle)
        W = 2*np.array([cang*U[0]+sang*V[0], cang*U[1]+sang*V[1], cang*U[2]+sang*V[2]])

    epsrgb = 0.

    setsat = False
    oksat = False
    didconverge = False
    dsat = 0.5
    iclipmin = int(cliprange[0]*(nlev-1)+0.5)
    iclipmax = int(cliprange[1]*(nlev-1)+0.5)
    kbest = 0
    for isat in range(iterSat+1):
        ii = 0
        d2clip = 1
        didclip = False
        while not didclip and ii <= 100:
            if iterSat <= 0:
                ksat = scalsat
            elif oksat:
                ksat = kbest*scalsat
            else:
                ksat = kbest+ii*dsat
            satar = ksat*np.linspace(minSat, maxSat, nlev)
            amp = satar * power_amp(fract_a, eps_amp, eps_pwr, pwr_amp)
            if k_ellipse > 0:
                kr = deform(angle, k_ellipse, alpha_ellipse, insectoid)
                amp = kr*amp

            # Tranformation to sRGB1 space
            if ColorModel >= 2: #UCS
                L = 100*fract_l
                a = 100*amp * np.cos(angle)
                b = 100*amp * np.sin(angle)
                Lab = np.array([L, a, b])
                srgb = cspace_converter("CAM02-UCS", "sRGB1")(Lab.T).T
            elif ColorModel == 1: #Standard CAM02
                JCh = np.array([100*fract_l, 100*amp, 180./pi*angle])
                srgb = cspace_converter("JCh", "sRGB1")(JCh.T).T
            else: #Classic cubehelix
                red = fract_l + amp * W[0]
                grn = fract_l + amp * W[1]
                blu = fract_l + amp * W[2]
                srgb = np.array([red, grn, blu])
            if trimWhite and ColorModel >= 1 and maxLight == 1:
                srgblin = cspace_converter("sRGB1", "sRGB1-linear")(srgb.T).T
                rgbw = srgblin[:, nlev-1]
                crgb = np.array([1, 1, 1]) / rgbw
                srgb = cspace_converter("sRGB1-linear", "sRGB1")(crgb*srgblin.T).T
                
            minrgb = np.amin(srgb[:, iclipmin:iclipmax])
            maxrgb = np.amax(srgb[:, iclipmin:iclipmax])
            d2clip = min(minrgb-epsrgb, 1-epsrgb-maxrgb)
            didclip = d2clip <= 0 or iterSat <= 0
            if didclip: kbest = max(0, ksat-dsat)
            ii += 1
        didconverge = dsat < epsSat
        oksat = didconverge or isat > iterSat or iterSat <= 0
        if setsat: break
        if oksat:
            setsat = True
        dsat /= 2
        #isat += 1
        if isat >= iterSat > 0: print('Warning: MaxSat algorithm did not converge.')
    auxpars.append(ksat)

    if verbose > 1: print('Using saturation scaling = {:.4f}'.format(ksat))

# optional endpoint tinting
    if do_tint:
        print('****tinting endpoints****')
        #print('islist0,1 = ',isinstance(pwrtint0, np.ndarray), isinstance(pwrtint1, np.ndarray))
        if isinstance(pwrtint0, (list, np.ndarray)) and isinstance(pwrtint1, (list, np.ndarray)):
            print('----Triple powers----')
            swtint0 = (1.-fract_full[:, np.newaxis])**np.array(pwrtint0)
            swtint1 = fract_full[:, np.newaxis]**np.array(pwrtint1)
            crgb0 = swtint0*np.array(rgbtint0) + (1.-swtint0)*np.array([1, 1, 1])
            crgb1 = swtint1*np.array(rgbtint1) + (1.-swtint1)*np.array([1, 1, 1])
            scalrgb0 = swtint0*np.array(scalrgb0) + (1.-swtint0)*np.array([1, 1, 1])
            scalrgb1 = swtint1*np.array(scalrgb1) + (1.-swtint1)*np.array([1, 1, 1])
        else: #assumes both pwrtint are scalars
            swtint0 = (1.-fract_full)**pwrtint0
            swtint1 = fract_full**pwrtint1
            crgb0 = swtint0[:, np.newaxis]*np.array(rgbtint0) + (1.-swtint0[:, np.newaxis])*np.array([1, 1, 1])
            crgb1 = swtint1[:, np.newaxis]*np.array(rgbtint1) + (1.-swtint1[:, np.newaxis])*np.array([1, 1, 1])
            scalrgb0 = swtint0[:, np.newaxis]*np.array(scalrgb0) + (1.-swtint0[:, np.newaxis])*np.array([1, 1, 1])
            scalrgb1 = swtint1[:, np.newaxis]*np.array(scalrgb1) + (1.-swtint1[:, np.newaxis])*np.array([1, 1, 1])
        if pwrtintw > 0:
            swtintw = swmixw*fract_full**pwrtintw
            crgb1 = swtintw[:, np.newaxis]*np.array([1, 1, 1]) + (1.-swtintw[:, np.newaxis])*crgb1
        crgb2 = crgb0*crgb1
        scalrgb2 = scalrgb0*scalrgb1
        srgblin = cspace_converter("sRGB1", "sRGB1-linear")(srgb.T).T
        if tintmode == 2:
            #print('gtint = ',gtint)
            rgbgray = (fract_full[:, np.newaxis]*np.array([1, 1, 1]))**gtint
            urgbgray = cspace_converter("sRGB1", "sRGB1-linear")(rgbgray).T
            drgb = (srgblin.T-urgbgray.T)
            #print(crgb2)
            #print('####')
            #print(scalrgb2)
            #print(drgb)
            #print(drgb*scalrgb2)
            tsrgblin = (drgb*scalrgb2 + urgbgray.T)
            #print(tsrgblin)
            srgb = cspace_converter("sRGB1-linear", "sRGB1")(crgb2*tsrgblin).T
        else:
            srgb = cspace_converter("sRGB1-linear", "sRGB1")(crgb2*srgblin.T).T

# optional color inverse
    if invmode == 1: #in sRGB
        srgb = np.sign(1.-srgb**ginv)*abs(1.-srgb**ginv)**(1./ginv)
    elif invmode >= 2: #in linear RGB
        srgblin = cspace_converter("sRGB1", "sRGB1-linear")(srgb)
        srgb = cspace_converter("sRGB1-linear", "sRGB1")(np.sign(1.-srgb**ginv)*abs(1.-srgblin**ginv)**(1./ginv))
        
# extract the R, G, B vectors
    red = srgb[0]
    grn = srgb[1]
    blu = srgb[2]

# find where RBB are outside the range [0,1], clip
    red[np.where((red > 1.))] = 1.
    grn[np.where((grn > 1.))] = 1.
    blu[np.where((blu > 1.))] = 1.

    red[np.where((red < 0.))] = 0.
    grn[np.where((grn < 0.))] = 0.
    blu[np.where((blu < 0.))] = 0.

# optional color reverse
    if reverse is True:
        red = red[::-1]
        blu = blu[::-1]
        grn = grn[::-1]

# put in to tuple & dictionary structures needed
    rr = []
    bb = []
    gg = []
    for k in range(0, int(nlev)):
        l = max(k-1, 0)
        rr.append((float(k) / (nlev - 1.), red[k], red[k]))
        bb.append((float(k) / (nlev - 1.), blu[k], blu[k]))
        gg.append((float(k) / (nlev - 1.), grn[k], grn[k]))

    cdict = {'red': rr, 'blue': bb, 'green': gg}
    return LSC('colorhelix_map', cdict, N=nlev)

def power_amp(x, eps_amp, eps_pwr, pwr_amp):
    if pwr_amp > 0:
        cdev = 0.25 * (4*x*(1.-x))**pwr_amp
    elif pwr_amp < 0:
        xf = 0.5+np.sign(x-0.5)*(0.5-x)
        x = 0.5*(1.-((2.*xf-1.)**eps_pwr+eps_amp**eps_pwr)**(1./eps_pwr))
        cdev = 0.5*(2*x)**(-pwr_amp)
    else:
        cdev = 1
    return cdev

def scurve(x, sw, pwr):
    x05 = 2*(x-0.5)
    if sw < 0 and pwr >= 0:
        xs = (1-abs(x05))**pwr
        xw = 1+np.sign(x-0.5)*(1-xs)
        xw = abs(sw)*xw+(1-abs(sw))*x*2
        xsc = xw/2
    else:
        xs = abs(x05)**abs(pwr) * np.sign(x05)
        xw = sw*xs + (1-sw)*x05
        xsc = 0.5*(xw+1)
    return xsc

def deform(angle, k_ellipse, alpha_ellipse, insectoid):
    if k_ellipse > 0:#safety catch
        ca = np.cos(angle+alpha_ellipse)
        sa = np.sin(angle+alpha_ellipse)
        if insectoid:
            kr = np.sqrt(ca**2+(k_ellipse*sa)**2)
        else:
            kr = k_ellipse/np.sqrt(sa**2+(k_ellipse*ca)**2)
    else:#default value
        kr = 1
    return kr
